# -*- coding: utf-8 -*-
"""
    mfe_saw.datasource
    ~~~~~~~~~~~~~

    This module imports into the mfe_saw core class to
    provide 'DevTree' and 'DataSource' objects.
"""
import csv
import ipaddress
import inspect
import json
import logging
import re
import sys
from itertools import chain
from io import StringIO
from functools import lru_cache


try:
    from mfe_saw.base import Base
    from mfe_saw.esm import ESM
    from mfe_saw.utils import dehexify
    from mfe_saw.exceptions import ESMException
except ImportError:
    from base import Base
    from esm import ESM
    from utils import dehexify
    from exceptions import ESMException

class DevTree(Base):
    """
    Interface to the ESM device tree.
    """
    _DevTree = []
    
    def __init__(self):
        """Coordinates assembly of the devtree"""
        super().__init__()
        if Base._baseurl == None:
            raise ESMException('ESM URL not set. Are you logged in?')
        
        if not DevTree._DevTree:
            self._esm = ESM()
            self._devtree = self._get_devtree()
            self._devtree = self.devtree_to_lod(self._devtree)
            self._devtree = self._add_parent_ids(self._devtree)
            self._parent_datasources = self.get_parent_datasources(self._devtree)
            self._clients = self.get_client_groups()
            self._clients = list(chain.from_iterable(self._clients))
            self._devtree = list(chain(self._devtree, self._clients))
            self._devtree = self._add_venmods(self._devtree)
            DevTree._DevTree = self._devtree

            
    def __len__(self):
        """
        Returns the count of devices in the device tree.
        """
        return len(DevTree._DevTree)
        
    def __iter__(self):
        """
        Returns:
            Generator with datasource objects.
        """
        for self._ds in DevTree._DevTree:
            yield self._ds

    def __contains__(self, term):
        """
        Returns:
            bool: True/False the name or IP matches the provided search term.
        """
        self._cterm = term
        if self.search(self._cterm, inc_hostname=True):
            return True
        else:
            return False

    def search(self, term, inc_hostname=False, zone_id=0):
        """
        Args:
            term (str): Datasource name, IP (or hostname) if inc_hostname 
            
            inc_hostname (bool): Search the hostname
            
            zone_id (int): Provide zone_id to limit search to a specific zone

        Returns:
            Datasource object that matches the provided search term or None.
        """
        self._term = term
        self._inc_hostname = inc_hostname

        self._search_fields = ['ds_ip', 'name']
        if self._inc_hostname:
            self._search_fields.append('hostname')
        
        self._found = [self._found_ds for self._field in self._search_fields
                      for self._found_ds in DevTree._DevTree
                      if self._found_ds[self._field].lower() == self._term.lower()]
                    
        if self._found:
            return self._found[0]
    
    @lru_cache(maxsize=None)
    def _get_devtree(self):
        """
        Returns:
            ESM device tree; raw, but ordered, string in need of parsing.
            Does not include client datasources.
        """
        self._method, self._data = self.get_params('get_devtree')
        self._resp = self.post(self._method, self._data)
        return dehexify(self._resp['ITEMS'])

    def devtree_to_lod(self, devtree):
        """
        Parse key fields from raw device strings.
        Return datasources as list of dicts
        """
        self._devtree = devtree
        self._devtree_io = StringIO(self._devtree)
        self._devtree_csv = csv.reader(self._devtree_io, delimiter=',')
        self._devtree_lod = []
        self._pid = None
        for self._row in self._devtree_csv:
            if len(self._row) == 0:
                continue
            if self._row[2] == "3":  # Client group datasource group containers
                self._row.pop(0)     # are fake datasources that seemingly have
                self._row.pop(0)     # two uneeded fields at the beginning.
            self._ds_fields = {'_dev_type': self._row[0],
                               'name': self._row[1],
                               'ds_id': self._row[2],
                               'enabled': self._row[15],
                               'ds_ip': self._row[27],
                               'hostname' : self._row[28],
                               'type_id': self._row[16],
                               'vendor': "",
                               'model': "",
                               'tz_id': "",
                               'date_order': "",
                               'port': "",
                               'syslog_tls': "",
                               'client_groups': self._row[29]
                               }
            self._devtree_lod.append(self._ds_fields)
        return self._devtree_lod

    def get_client_groups(self):
        """
        Retrieve client lists from each parent group
        
        Args:
            ds_id (str): Parent ds_id(s) are collected on init
            ftoken (str): Set and used after requesting clients for ds_id
        
        Returns:
            List of dicts representing all of the client data sources 
        """
        self.client_lod = []
        for self.parent in self._parent_datasources:
            self.ds_id = self.parent['ds_id']
            self._resp = self.find_client_group(self.ds_id)
            self._ftoken = self._resp['FTOKEN']
            self._resp = self.get_file(self._ftoken)
            self.client_dict = self.clients_to_lod(self._resp, self.ds_id)
            self.client_lod.append(self.client_dict)
        return self.client_lod

    def clients_to_lod(self, clients, ds_id):
        """
        Parse key fields from 'DS_GETDSCLIENTLIST'.
        Return clients as list of dicts
        """
        self._clients = clients
        self._ds_id = ds_id

        self._clients_io = StringIO(self._clients)
        self._clients_csv = csv.reader(self._clients_io, delimiter=',')

        self._parsed_clients = []
        for self._row in self._clients_csv:
            if len(self._row) < 2:
                continue
            if self._row[2] == "3":
                self._row.pop(0)
                self._row.pop(0)

            self._ds_fields = {'_dev_type': "0",
                              'name': self._row[1],
                              'id': self._row[0],
                              'enabled': self._row[2],
                              'ds_ip': self._row[3],
                              'hostname' : self._row[4],
                              'type_id': self._row[5],
                              'vendor': self._row[6],
                              'model': self._row[7],
                              'tz_id': self._row[8],
                              'date_order': self._row[9],
                              'port': self._row[11],
                              'syslog_tls': self._row[12],
                              'client_groups': "0",
                              'parent_id': self._ds_id
                              }
            self._parsed_clients.append(self._ds_fields)
        return self._parsed_clients

    def get_parent_datasources(self, ds_summary):
        """
        Parse dict for parent datasources
        Returns dict
        """
        self.ds_summary = ds_summary
        self.ds_parents = []
        for self.ds in self.ds_summary:
            if self.ds['_dev_type'] == "3" and int(self.ds['client_groups']) > 0:
                self.ds_parents.append(self.ds)
        return self.ds_parents

    def find_client_group(self, group_id):
        """
        Finds client group
        
        Args:
            DSID (str): Parent datasource ID set to self._ds_id
        
        Returns:
            Response dict with FTOKEN required for next step: 
        
        """
        self.group_id = group_id
        self._method, self._data = self.get_params('find_client_group')
        self._resp = self.post(self._method, self._data)
        return self._resp

    def get_file(self, ftoken):
        """
        Exchanges token for file
        
        Args:
            ftoken (str): instance name set by 
        
        """
        self.ftoken = ftoken
        self._method, self._data = self.get_params('get_file')
        self._resp = self.post(self._method, self._data)
        self._resp = dehexify(self._resp['DATA'])
        return self._resp

    def _add_venmods(self, devtree_lod):
        """
        Populates vendor/model fields for any datasources 
        """
        self._devtree_lod = devtree_lod
        for self._ds in self._devtree_lod:
            if not self._ds['vendor'] and self._ds['_dev_type'] == '3':
                (self._ds['vendor'], 
                self._ds['model']) = self._esm.type_id_to_venmod(self._ds['type_id'])
        return self._devtree_lod
    
    def _add_parent_ids(self, devtree_lod):
        """
        """
        self._devtree_lod = devtree_lod
        for self._ds in self._devtree_lod:
            if self._ds['_dev_type'] == "2":
                self._pid = self._ds['ds_id']
            if self._pid:
                self._ds['parent_id'] = self._pid
        return self._devtree_lod

                
class DataSource(Base):
    """
    A DataSource object represents a validated datasource configuration.
    This object represents current datasources as well as new datasources 
    to be added to the tree. 
    
    The kwargs required to initialize this class are detailed in __init__.
    """
    
    def __init__(self, dsconf):
        """
        Inits the datasource
        
        Args: 
            dsconf (dict of str: str)
            
            The dict may hold any number of valid datasource args, 
            but at a mininum, the following arguments are required:
            
            name (str): datasource name
            type_id (str): datasource type_id
            parent_id (str): datasource parent_id
            ip (str): unique IP address of datasource OR
            hostname (str): unique hostname 
            **kwargs: optional props
            
            Note:
            Both hostname and ip can be set, but at least one of them
            must be set.
            
        """
        super().__init__()
        if Base._baseurl == None:
            raise ESMException('ESM URL not set. Are you logged in?')
        self._esm = ESM()
        self._devtree = DevTree()
        self._ds_conf = dsconf
        self.ds_id = None
        self.child_enabled = "false"
        self.child_count = "0"
        self.child_type = "0"
        self.zone_id = "0"
        self.url = None
        self.enabled = 'true'
        self.idm_id = "0"
        self.hostname = None
        self.tz_id = None
        self.dorder = None
        self.maskflag = None
        self.port = None
        self.syslog_tls = None
        self.vendor = None
        self.model = None
        self.client_groups = None
        self.__dict__.update(self._ds_conf)
        self._prop = None
        self._pval = None

    def delete(self):
        """
        Deletes a datasource
        
        Args:
            ds_id (str). DataSource ID
            rec_id (str). Receiver ID / DataSource parent_id
            
        Warning:
            This really does delete the datasource and ALL data
            ever collected for that datasource.
        
        Returns:
            None
        
        Raises:
            ESMException: If the datasource to be deleted is 
                still in the tree after being deleted an Exception 
                will be raised.
        """
        print(self._ds_details())
        self._method, self._data = self.get_params('del_ds')
        self._resp = self.post(self._method, self._data)
        try:
            self._ds_details()
        except ESMDataSourceNotFound:
            print("WOWOWWOWA")
        
    def _ds_details(self):
        """
        Queries the ESM for datasource details
        
        Returns:
            dict (str, str) with some subdicts 
        
        Warning:
            Don't create a situation where this gets called for every
            datasource as it will not scale.
        """
        self._method, self._data = self.get_params('ds_details')
        return self.post(self._method, self._data)
        
            
    def add(self, client=False):
        """
        Adds a datasource
        
        Returns:
            None 
        
        Raises:
            ESMException: Will be raised if trying to add a duplicate datasource.
            
        """
        
        if self._devtree.search(self.name, zone_id=self.zone_id):
            raise ESMException('Datasource name already exists. '
                                'Cannot add datasource: {}'.format(self.name))
        if self._devtree.search(self.ds_ip, zone_id=self.zone_id):
            raise ESMException('Datasource IP already exists.' 
                                'Cannot add datasource: {}'.format(self.name))
            
        if client:
            self._method, self._data = self.get_params('add_client')
        else:
            self._method, self._data = self.get_params('add_parent')
        self._resp = self.post(self._method, self._data)
        self.parent_id = self._resp.get('id')
                                     
    def __repr__(self):
        """
        Dumps the datasource settings in json
        
        Returns:
            str: Datasource attributes as JSON
        """        
        return json.dumps(self.props())
    
    def props(self):
        """
        Dumps the datasource settings
        
        Returns:
            str: Datasource attributes as JSON
        """        
        return {self._prop: self._pval
            for self._prop, self._pval in self.__dict__.items()
            if not self._prop.startswith('_')}


    @staticmethod
    def valid_ip(ipaddr):
        """
        Validates IPv4/v6 address or raises ValueError.

        Args:
            ipaddr (str): IP address

        Returns:
            True if valid, False if not.
            
        Raises:
            ValueError: It's the wrong value if it's not valid.
        """
        try:
            ipaddr = str(ipaddress.ip_address(ipaddr))
            return True
        except ValueError:
            return False



# _dev_typeesm.time()
# _dev_typeesm.etime()
# _dev_typeesm.status()
# _dev_typeesm.disks()
# _dev_typeesm.ram()
# _dev_typeesm.backup_status()
# _dev_typeesm.callhome()
# _dev_typeesm.rules_status()
# _dev_typeesm._get_timezones()
# _dev_typeesm.tz_offsets()
# _dev_typeesm.timezones()
# _dev_typeesm.tz_name_to_id('Darwin')
# _dev_typeesm.tz_id_to_name('1')
# _dev_typeesm.type_id_to_venmod('65')
# _dev_typeesm.venmod_to_type_id('UNIX', 'Linux')
# _dev_typeesm._get_ds_types()
# _dev_typeesm.recs()



# -*- coding: utf-8 -*-
"""
    mfe_saw

"""

import time
from functools import wraps

def dehexify(data):
    """
    A URL and Hexadecimal Decoding Library.

    Credit: Larry Dewey
    """

    hexen = {
        '\x1c': ',',  # Replacing Device Control 1 with a comma.
        '\x11': ',',  # Replacing Device Control 2 with a new line.
        '\x12': '\n',  # Space
        '\x22': '"',  # Double Quotes
        '\x23': '#',  # Number Symbol
        '\x27': '\'',  # Single Quote
        '\x28': '(',  # Open Parenthesis
        '\x29': ')',  # Close Parenthesis
        '\x2b': '+',  # Plus Symbol
        '\x2d': '-',  # Hyphen Symbol
        '\x2e': '.',  # Period, dot, or full stop.
        '\x2f': '/',  # Forward Slash or divide symbol.
        '\x7c': '|',  # Vertical bar or pipe.
    }

    uri = {
        '%11': ',',  # Replacing Device Control 1 with a comma.
        '%12': '\n',  # Replacing Device Control 2 with a new line.
        '%20': ' ',  # Space
        '%22': '"',  # Double Quotes
        '%23': '#',  # Number Symbol
        '%27': '\'',  # Single Quote
        '%28': '(',  # Open Parenthesis
        '%29': ')',  # Close Parenthesis
        '%2B': '+',  # Plus Symbol
        '%2D': '-',  # Hyphen Symbol
        '%2E': '.',  # Period, dot, or full stop.
        '%2F': '/',  # Forward Slash or divide symbol.
        '%7C': '|',  # Vertical bar or pipe.
    }

    for (enc, dec) in hexen.items():
        data = data.replace(enc, dec)

    for (enc, dec) in uri.items():
        data = data.replace(enc, dec)

    return data


def timethis(func):
    """
    Decorator that reports the execution time.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        """Wrapper"""
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        print(func.__name__, end-start)
        return result
    return wrapper



# -*- coding: utf-8 -*-

"""Console script for mfe_saw"""

import argparse

try:
    from mfe_saw.esm import ESM
    from mfe_saw.datasource import Datasource, DevTree
except ModuleNotFoundError:
    from esm import ESM
    from datasource import Datasource, DevTree

    
class Args(object):

    def __init__(self, args):
        self.log_levels = ['quiet', 'error', 'warning', 'info', 'debug']
        self.output_formats = ['json', 'csv', 'raw', 'word']
        self.formatter_class = argparse.RawDescriptionHelpFormatter
        self.parser = argparse.ArgumentParser(
                formatter_class=self.formatter_class,
                description='McAfee SIEM API Wrapper'
            )
        self.args = args

        self.parser.add_argument('-a', '--add', 
                                 action='store_true', dest='add', default=None,
                                 help='Scan <dsdir> for new datasource files')

        self.parser.add_argument('-s', '--search', 
                                 dest='search', nargs='?', default=None, metavar='term',
                                 help='Search for datasource name, hostname, or IP.')
        
        self.parser.add_argument('-d', '--dsconf'
                                 action='store_true', dest='dssum', default=None,
                                 help='Print datasource configs.')

        self.parser.add_argument('-o', 
                                 choices=self.output_formats, dest='output',
                                         default='default', metavar='[format]',
                                 help=('Set output format of results. Default:'
                                        'columns with borders. '
                                        'Options: json, csv, raw, and word')
                                       
        self.parser.add_argument('-w', 
                                 action='store_true', dest='write',default=None,
                                 help='Write output to file.')
                                 
        self.parser.add_argument('-v', '--version', 
                                 action='version',
                                 help='Show version',
                                 version='%(prog)s {}'.format(__version__))

        self.parser.add_argument('-l', '--level',
                                 default=None, dest='level',
                                 choices=self.log_levels, metavar='',
                                 help='Logging output level. Default: warning')
        
        self.parser.add_argument('-c', '--config',
                                 default=None, dest='cfgfile', metavar='',
                                 help='Path to config file. Default: config.ini')                                 

        self.pargs = self.parser.parse_args()
            
    def get_args(self):
        return self.pargs

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
    mfe_saw __main__
    ~~~~~~~~~~~~~


"""
import json
import requests
import sys
try:
    from mfe_saw.esm import ESM
    from mfe_saw.datasource import DataSource
    from mfe_saw.utils import timethis
except ImportError:
    from esm import ESM
    from datasource import DataSource, DevTree
    from utils import timethis

@timethis
def main():
    """
    Main function
    """
    
main()





# -*- coding: utf-8 -*-
"""
    mfe_saw ESM Class

"""
import json
from functools import lru_cache

try:
    from mfe_saw.base import Base
except ImportError:
    from base import Base

class ESM(Base):
    """
    ESM class
    """
    def __init__(self):
        """
        Args:
            host (str): IP or hostname of ESM.

        Returns:
            obj. ESM object
        """
        super().__init__()

    def version(self):
        """
        Returns:
            str. ESM short version.

        Example:
            '10.0.2'
        """
        return self.buildstamp().split()[0]

    def buildstamp(self):
        """
        Returns:
            str. ESM buildstamp.

        Example:
            '10.0.2 20170516001031'
        """
        return self.post('essmgtGetBuildStamp')['buildStamp']

    def time(self):
        """
        Returns:
            str. ESM time (GMT).

        Example:
            '2017-07-06T12:21:59.0+0000'
        """
        self._esmtime = self.post("essmgtGetESSTime")
        return self._esmtime['value']


    def status(self):
        """
        Returns:
            dict. ESM stats.
            including:
                - processor status
                - hdd status
                - ram status
                - rule update status
                - backup status
                - list of top level devices
        Other functions exist to return subsets of this data also.
        """
        return self.post("sysGetSysInfo")

    def disks(self):
        """
        Returns:
            str. ESM disks and utilization.

        Example:
            'sda3     Size:  491GB, Used:   55GB(12%), Available:  413GB, Mount: /'
        """
        return self.status()['hdd']

    def ram(self):
        """
        Returns:
            str. ESM ram and utilization.

        Example:
            'Avail: 7977MB, Used: 7857MB, Free: 119MB'
        """
        return self.status()['ram']

    def backup_status(self):
        """
        Returns:
            dict. Backup status and timestamps.

        Example:
            {'autoBackupEnabled': True,
                'autoBackupDay': 7,
                'autoBackupHour': 0,
                'backupLastTime': '07/03/2017 08:59:36',
                'backupNextTime': '07/10/2017 08:59'}
        """
        self._fields = ['autoBackupEnabled',
                        'autoBackupDay',
                        'autoBackupHour',
                        'autoBackupHour',
                        'backupNextTime']

        return {self.key: self.val for self.key, self.val in self.status().items()
                if self.key in self._fields}

    def callhome(self):
        """
        Returns:
            bool. True/False if there is currently a callhome connection
        """
        self._callhome_ip = self.status()['callHomeIp']
        if self._callhome_ip:
            return True

    def rules_status(self):
        """
        Returns:
            dict. Rules autocheck status and timestamps.

        Example:
        { 'rulesAndSoftwareCheckEnabled': True
          'rulesAndSoftLastCheck': '07/06/2017 10:28:43',
          'rulesAndSoftNextCheck': '07/06/2017 22:28:43',}

        """
        self._fields = ['rulesAndSoftwareCheckEnabled',
                        'rulesAndSoftLastCheck',
                        'rulesAndSoftNextCheck']
        return {self.key: self.val for self.key, self.val in self.status().items()
                if self.key in self._fields}

    @lru_cache(maxsize=None)   
    def _get_timezones(self):
        """
        Gets list of timezones from the ESM.
        
        Returns:
            str. Raw return string from ESM including 
        """
        return self.post('userGetTimeZones')

        
    def tz_offsets(self):
        """
        Builds table of ESM timezones including offsets.
        
        Returns:
            list. List of timezone tuples (name, id, offset)
            
        Example:
            [(1, 'Midway Island, Samoa', '-11:00'),
             (2, 'Hawaii', '-10:00'),
             ...
            ]
        """
        self.tz_resp = self._get_timezones()
        return [(self.tz['id']['value'], self.tz['name'], self.tz['offset']) 
                  for self.tz in self.tz_resp]
                   
        
    def timezones(self):
        """
        Builds table of ESM timezones and names only. No offsets.
        
        Returns:
            dict. {timezone_id, timezone_name}
        """
        self.tz_resp = self._get_timezones()
        self.tz_table = {str(self.tz['id']['value']): self.tz['name']
                            for self.tz in self.tz_resp}
        return self.tz_table

    def tz_name_to_id(self, tz_name):
        """
        Args:
            tz_name (str): Case sensitive, exact match timezone name
            
        Returns:
            str. Timezone id or None if there is no match
        """
        self.tz_reverse = {self.tz_name: self.tz_id 
                            for self.tz_id, self.tz_name in self.timezones().items()}
        try:
            return self.tz_reverse[tz_name]
        except KeyError:
            return None
    
    def tz_id_to_name(self, tz_id):
        """
        Args:
            td_id (str): Numerical string (Currently 1-74)
        
        Returns:
            str. Timezone name or None if there is no match
        """
        try:
            return self.timezones()[tz_id]
        except KeyError:
            return None
    
    def type_id_to_venmod(self, type_id):
        """
        Args:
            type_id (str): Numerical string 
        
        Returns:
            tuple. (vendor, model) or None if there is no match
        """
        self.type_id = type_id
        self.ds_types = self._get_ds_types()
        for self.venmod in self.ds_types:
            if str(self.venmod[0]) == str(self.type_id):
                return (self.venmod[1], self.venmod[2])

    def venmod_to_type_id(self, vendor, model):
        """
        Args:
            vendor (str): Exact vendor string including puncuation
            model (str): Exact vendor string including puncuation
        
        Returns:
            str. Matching type_id or None if there is no match
        """
        self.vendor = vendor
        self.model = model
        self.ds_types = self._get_ds_types()
        for self.venmod in self.ds_types:
            if self.vendor == self.venmod[1]:
                if self.model == self.venmod[2]:
                    return str(self.venmod[0])
        
     
    @lru_cache(maxsize=None)   
    def _get_ds_types(self):
        """
        Retrieves device table from ESM
                    
        Returns:
            list. of tuples output from callback: _format_ds_types()

        Note:
            rec_id (str): self.rec_id assigned in method
        """
        self.rec_id = self.recs()[0][1]
        self.method, self.data = self.get_params('get_dstypes')
        self.venmods = self.post(self.method, self.data, self._format_ds_types)
        return self.venmods
                    
    def _format_ds_types(self, venmods):
        """
        Callback to create type_id/vendor/model table
        
        Args:
            venmods (obj): request object from _get_ds_types
        
        Returns:
            list. of tuples 
                
           [(542, 'McAfee', 'SaaS Email Protection')
            (326, 'McAfee', 'Web Gateway')
            (406, 'Microsoft', 'ACS - SQL Pull')
            (491, 'Microsoft', 'Endpoint Protection - SQL Pull')
            (348, 'Microsoft', 'Exchange')]

        Note: 
            This is a callback for _get_ds_types.

        """
        self._venmods = venmods
        return [(self._mod['id']['id'], self._ven['name'], self._mod['name'],)
                    for self._ven in self._venmods['vendors']
                    for self._mod in self._ven['models']]

    @lru_cache(maxsize=None)    
    def recs(self):
        """
        
        """
        self.method, self.data = self.get_params('get_recs')
        self._rec_list = self.post(self.method, self.data)
        return [(self._rec['name'], self._rec['id']['id']) 
                  for self._rec in self._rec_list]



# -*- coding: utf-8 -*-
"""
    mfe_saw

"""

import ast
import base64
import json
import re
import urllib.parse as urlparse
from concurrent.futures import ThreadPoolExecutor

import requests
import urllib3

try:
    from mfe_saw.params import PARAMS
    from mfe_saw.exceptions import ESMException, ESMDataSourceNotFound
except ImportError:
    from params import PARAMS
    from exceptions import ESMException, ESMDataSourceNotFound

class Base(object):
    """
    The Base class for mfe_saw objects
    """
    _headers = {'Content-Type': 'application/json'}
    _baseurl = None
    _basepriv = None
    _max_workers = 10
    _ssl_verify = False
    _params = PARAMS
    
    def __init__(self, **kwargs):
        """
        Base Class for mfe_saw objects.

        """
        self._kwargs = kwargs

        self._url = None
        self._data = None
        self._uri = None
        self._resp = None
        self._host = None
        self._user = None
        self._passwd = None
        self._username = None
        self._password = None
        self._cmd = None
        self._future = None
        self._result = None
        self._method = None
        self._name = None

        if not self._ssl_verify:
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        self._ex = ThreadPoolExecutor(max_workers=Base._max_workers,)

    def login(self, host, user, passwd):
        """
        The login method
        """
        self._host = host
        self._user = user
        self._passwd = passwd

        Base._baseurl = 'https://{}/rs/esm/'.format(self._host)
        Base._basepriv = 'https://{}/ess'.format(self._host)

        self._username = base64.b64encode(self._user.encode('utf-8')).decode()
        self._password = base64.b64encode(self._passwd.encode('utf-8')).decode()
        del self._passwd
        self._url = Base._baseurl + 'login'
        self._method, self._data = self.get_params('login')
        self._resp = self.post(self._method, self._data, raw=True)
        try:
            Base._headers['Cookie'] = self._resp.headers.get('Set-Cookie')
            Base._headers['X-Xsrf-Token'] = self._resp.headers.get('Xsrf-Token')
        except AttributeError:
            raise ESMAuthError()
            
    def get_params(self, method):
        """
        Look up parameters in params dict
        """
        self._method = method
        self._method, self.data = self._params.get(self._method)
        self._data = self.data % self.__dict__
        self._data = ast.literal_eval(''.join(self._data.split()))
        return self._method, self._data

    @staticmethod
    def _format_params(cmd, **params):
        """
        Format private API call
        """
        params = {k: v for k, v in params.items() if v is not None}
        params = '%14'.join([k + '%13' + v + '%13' for (k, v) in params.items()])
        params = 'Request=API%13' + cmd + '%13%14' + params + '%14'
        return params

    @staticmethod
    def _format_priv_resp(resp):
        """
        Format response from private API
        """
        resp = re.search('Response=(.*)', resp).group(1)
        resp = resp.replace('%14', ' ')
        pairs = resp.split()
        formatted = {}
        for pair in pairs:
            pair = pair.replace('%13', ' ')
            pair = pair.split()
            key = pair[0]
            if key == 'ITEMS':
                value = pair[-1]
            else:
                value = urlparse.unquote(pair[-1])
            formatted[key] = value
        return formatted

    def post(self, method, data=None, callback=None, raw=False):
        """
        Wrapper around _post method
        """
        self._method = method
        self._data = data
        self._callback = callback
        self._raw = raw
        self._url = Base._baseurl + self._method
        
        if self._method == self._method.upper():
            self._url = Base._basepriv
            self._data = self._format_params(self._method, **self._data)
        else:
            self._url = Base._baseurl + self._method
            if self._data:
                try:
                    self._data = json.dumps(self._data)
                except json.JSONDecodeError:
                    raise ESMParamsError()

        self._future = self._ex.submit(self._post, url=self._url,
                                     data=self._data,
                                     headers=self._headers,
                                     verify=self._ssl_verify)
        self._resp = self._future.result()

        if self._raw:
            return self._resp

        if 200 <= self._resp.status_code <= 300:
            try:
                self._resp = self._resp.json()
                self._resp = self._resp.get('return')
            except json.decoder.JSONDecodeError:
                self._resp = self._resp.text
            if self._method == self._method.upper():
                self._resp = self._format_priv_resp(self._resp)
            if self._callback:
                self._resp = self._callback(self._resp)
            return self._resp
        elif self._resp.status_code == 400: 
            if self._resp.text.startswith('Error deserializing EsmDataSourceDetail'):
                raise ESMDataSourceNotFound
        

    def _post(self, url, data=None, headers=None, verify=False):
        """
        Method that actually kicks off the HTTP client.
        
        Args:
            url (str): URL to send the post to.
            data (str): Any payload data for the post.
            headers (str): http headers that hold cookie data after 
                            authentication.
            verify (bool): SSL cerificate verification 
        
        Returns:
            Requests Response object
        """
        self._url = url
        self._data = data
        self._headers = headers
        self._verify = verify
        return requests.post(self._url, 
                                    data=self._data, 
                                    headers=self._headers, 
                                    verify=self._verify)


# -*- coding: utf-8 -*-

import os

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


with open('README.rst') as readme_file:
    readme = readme_file.read()

test_requirements = ['pytest', 'tox']
requirements = ['requests', 'prettytable']
    
    
setup(
    name='mfe_saw',
    version='0.0.2',
    description="McAfee SIEM API Wrapper (MFE_SAW) for McAfee ESM 10.x+",
    author="Andy Walden",
    author_email='aw@krakencodes.com',
    url='https://github.com/andywalden/mfe_saw',
    packages=['mfe_saw'],
    package_dir={'mfe_saw': 'mfe_saw'},
    include_package_data=True,
    install_requires=requirements,
    license="ISC",
    keywords='mfe_saw',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only'],
    test_suite='tests',
    tests_require=test_requirements,
    python_requires='>=3')



# -*- coding: utf-8 -*-
"""
    mfe_saw.params
    ~~~~~~~~~~~~~

    This module imports a dict into the mfe_saw core class to provide
    a central place to aggregate methods and parameters. The params
    are stored as docstrings to support string replacement.

    Args:
        method (str): Dict key associated with desired function
        Use normal dict access, ['method'], or .pop('method')

    Returns:
        tuple: (string, string)

        The first string is the method name that is actually used as
        the URI or passed to the ESM. The second string is the params
        required for that method. Some params require variables be
        interpolated as documented in the Attributes.

    Example:
        method, params = params['login'].format(username, password)

    Attributes:
        login: Function to login
            vars:
                username
                password
            callback vars:
                Cookie
                X-Xsrf-Token

        devtree: Get top level device tree string

        client_grp: Get clients for a specific group
            vars:
                id
            callback vars:
                ftoken

        results: Get results from earlier call
            vars:
                ftoken

"""

PARAMS = {
    'login': ("login",
              """{'username': '%(_username)s',
                 'password' : '%(_password)s',
                 'locale': 'en_US',
                 'os': 'Win32'}
                 """),

    'get_devtree': ("GRP_GETVIRTUALGROUPIPSLISTDATA",
                    """{'ITEMS': '#{DC1 + DC2}',
                    'DID': '1',
                    'HD': 'F',
                    'NS': '0'}
                    """),

    'find_client_group': ("DS_GETDSCLIENTLIST",
                          """{'DSID': '%(group_id)s',
                            'SEARCH': ''}
                         """),

    'get_file': ("MISC_READFILE",
                 """{'FNAME': '%(ftoken)s',
                 'SPOS': '0',
                 'NBYTES': '0'}
                 """),

    'map_dtree': ("map_dtree",
                  """{'dev_type': '%(dev_type)s',
                  'name': '%(ds_name)s',
                  'ds_id': '%(ds_id)s',
                  'enabled': '%(enabled)s',
                  'ds_ip': '%(ds_ip)s',
                  'hostname' : '%(hostname)s',
                  'typeID': '%(type_id)s',
                  'vendor': '',
                  'model': '',
                  'tz_id': '',
                  'date_order': '',
                  'port': '',
                  'syslog_tls': '',
                  'client_groups': '%(client_groups)s'
                  }
                  """),
                 
    'add_parent': ("dsAddDataSource", 
                """ {'datasource': {
                        'parentId': {'id': '%(parent_id)s'},
                        'name': '%(name)s',
                        'id': {'id': '%(ds_id)s'},
                        'typeId': {'id': '%(type_id)s'},
                        'childEnabled': '%(child_enabled)s',
                        'childCount': '%(child_count)s',
                        'childType': '%(child_type)s',
                        'ipAddress': '%(ds_ip)s',
                        'zoneId': '%(zone_id)s',
                        'url': '%(url)s',
                        'enabled': '%(enabled)s',
                        'idmId': '%(idm_id)s'
                    }}"""),

    'add_client': ("DS_ADDDSCLIENT", 
                     """{'PID': '%(parent_id)s',
                     'NAME': '%(name)s',
                     'ENABLED': '%(enabled)s',
                     'IP': '%(ds_ip)s',
                     'HOST': '%(hostname)s',
                     'TYPE': '%(type_id)s',
                     'TZID': '%(tz_id)s',
                     'DORDER': '%(dorder)s',
                     'MASKFLAG': '%(maskflag)s',
                     'PORT': '%(port)s',
                     'USETLS': '%(syslog_tls)s'
                    }"""),
                    
    'get_recs': ("devGetDeviceList?filterByRights=false",
                     """{'types': ['RECEIVER']}
                     """),

    'get_dstypes': ("dsGetDataSourceTypes",
                     """{'receiverId': {'id': '%(rec_id)s'}
                        }
                     """),
                     
    'del_ds': ("dsDeleteDataSource",
                """{'receiverId': {'id': '%(parent_id)s'},
                    'datasourceId': {'id': '%(ds_id)s'}}
                 """),
                 
    'ds_last_times': ("QRY_GETDEVICELASTALERTTIME",
                      """
                      """),
                      
    'ds_by_type': ("QRY_GETDEVICECOUNTBYTYPE",
                      """
                      """),

   '_dev_types':  {'2', 'ERC',
                    '3', 'datasource',
                    '4', 'Database Event Monitor (DBM)',
                    '5', 'DBM Database',
                    '7', 'Policy Auditor',
                    '10', 'Application Data Monitor (ADM)',
                    '12', 'ELM'
                    '14', 'Local ESM',
                    '15', 'Advanced Correlation Engine (ACE)',
                    '17', 'Score-based Correlation',
                    '19', 'McAfee ePolicy Orchestrator (ePO)',
                    '20', 'EPO',
                    '21', 'McAfee Network Security Manager (NSP)',
                    '23', 'NSP Port',
                    '24', 'MVM',
                    '25', 'Enterprise Log Search (ELS)',
                    '254', 'client_group',
                    },
                    
    'ds_details': ("dsGetDataSourceDetail",
                    """{'datasourceId': 
                        {'id': '%(ds_id)s'}}
                    """)
}